/*
     RAEAN-1b: NormsAgent.java
  Copyright (C) 2004  Universidad de Burgos

     This file is part of RAEAN (Re-implementation of Axelrod's "Evolutionary approach to norms"),
 an agent-based model designed to replicate and extend Prof. Axelrods
 agent-based model described in the paper "An evolutionary approach to
 norms", published in American Political Science Review 80, n4,
 December 1996: 1095-1111, and reprinted in chapter 3 (Promoting
 norms) of Robert M. Axelrods book "The Complexity of Cooperation.
 Agent-Based Models of Competition and Collaboration", Princeton
 University Press. Princeton, New Jersey 1997.

 The code is written in Java 2, and is known to work using RePast-2.2
 libraries (http://repast.sourceforge.net/) and JDK 1.4.2_01-b06
 distribution on a PC running Windows 2000, NT and XP.

     RAEAN is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation; either version 2 of the License, or
     (at your option) any later version.

     RAEAN is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details. (LICENCE file in
     this directory.)

     You should have received a copy of the GNU General Public License
     along with this program; if not, write to the Free Software
     Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

     Contact information:
       Jos M. Galn,
       Universidad de Burgos, Burgos, Spain
       jmgalan@ubu.es
 */





import cern.jet.random.Uniform;

/**
 * <p>Title: Re-implementation of Axelrod's "Evolutionary approach to norms"</p>
 * <p>Description: Reimplementation by Jos M. Galn & Luis R. Izquierdo</p>
 * <p>Copyright: Universidad de Burgos</p>
 * <p>Company: Universidad de Burgos & Macauley Institute & INSISOC Group</p>
 * @author Jos Manuel Galn & Luis R. Izquierdo (in alphabetic order)
 * @version 1.0
 */
public class NormsAgent {
  /**
   * ArrayDimension determines the number of bits representing each of the characteristics of the agents
   */
  public int arrayDimension;

  /**
   * bit-string boldness of an individual agent
   */
  private int[] boldness;

  /**
   * bit-string vengefulness of an individual agent
   */
  private int[] vengefulness;

  /**
   * bit-string metavengefulness of an individual agent
   */
  private int[] metaVengefulness;

  /**
   * integer boldness of an individual agent
   */
  private int boldnessDecimal = 0;

  /**
   * integer vengefulness of an individual agent
   */
  private int vengefulnessDecimal = 0;

  /**
   * integer metavengefulness of an individual agent
   */
  private int metaVengefulnessDecimal = 0;

  /**
   * probability of being observed (S) in each round
   */
  private double variable_S = 0;

  /**
   * score of each round
   */
  public double[] score;

  /**
   * Total score in a generation
   */
  public double totalScoreInTheRounds = 0;

  /**
   * Defection or not in each round
   */
  private boolean defection = false;

  /**
   * Constructor for Random initialization
   * @param dimension int
   * @param rounds int
   */
  public NormsAgent(int dimension, int rounds) {

    arrayDimension = dimension;

//random features of the agents
    boldness = new int[arrayDimension];
    vengefulness = new int[arrayDimension];
    metaVengefulness = new int[arrayDimension];
    score = new double[rounds];

    for (int i = 0; i < score.length; i++) {
      score[i] = 0;
    }

    setRandomValuesInABitVector(boldness, arrayDimension);
    setRandomValuesInABitVector(vengefulness, arrayDimension);
    setRandomValuesInABitVector(metaVengefulness, arrayDimension);

    boldnessDecimal = setBitValuesInDecimalValues(boldness);
    vengefulnessDecimal = setBitValuesInDecimalValues(vengefulness);
    metaVengefulnessDecimal = setBitValuesInDecimalValues(metaVengefulness);

  }

  /**
   * Constructor for Custom initialization
   * @param dimension int
   * @param rounds int
   * @param initialBoldnessAverage double
   * @param initialVengefulnessAverage double
   * @param initialMetaVengefulnessAverage double
   */
  public NormsAgent(int dimension, int rounds, double initialBoldnessAverage,
                    double initialVengefulnessAverage,
                    double initialMetaVengefulnessAverage) {
    arrayDimension = dimension;

    boldness = new int[arrayDimension];
    vengefulness = new int[arrayDimension];
    metaVengefulness = new int[arrayDimension];
    score = new double[rounds];

    for (int i = 0; i < score.length; i++) {
      score[i] = 0;
    }

    setFixedValuesInABitVector(initialBoldnessAverage, boldness, arrayDimension);
    setFixedValuesInABitVector(initialVengefulnessAverage, vengefulness,
                               arrayDimension);
    setFixedValuesInABitVector(initialMetaVengefulnessAverage, metaVengefulness,
                               arrayDimension);

    boldnessDecimal = setBitValuesInDecimalValues(boldness);
    vengefulnessDecimal = setBitValuesInDecimalValues(vengefulness);
    metaVengefulnessDecimal = setBitValuesInDecimalValues(metaVengefulness);

  }

  /**
   * Mutation of an agent
   * @param parent NormsAgent
   * @param rateMutation double
   */
  public NormsAgent(NormsAgent parent, double rateMutation) {

    this.arrayDimension = parent.arrayDimension;
    //Create
    boldness = new int[arrayDimension];
    vengefulness = new int[arrayDimension];
    metaVengefulness = new int[arrayDimension];
    score = new double[parent.score.length];

    System.arraycopy(parent.boldness, 0, this.boldness, 0, arrayDimension);
    System.arraycopy(parent.vengefulness, 0, this.vengefulness, 0,
                     arrayDimension);
    System.arraycopy(parent.metaVengefulness, 0, this.metaVengefulness, 0,
                     arrayDimension);

    /*=======================MUTATION==========================================*/

    for (int i = 0; i < this.boldness.length; i++) {
      if (Uniform.staticNextDoubleFromTo(0, 1) <= rateMutation) { //Peligro de error de punto flotante
        if (boldness[i] == 1) {
          boldness[i] = 0;
        }
        else {
          boldness[i] = 1;
        }
      }
    }

    for (int i = 0; i < this.vengefulness.length; i++) {
      if (Uniform.staticNextDoubleFromTo(0, 1) <= rateMutation) { //Peligro de error de punto flotante
        if (vengefulness[i] == 1) {
          vengefulness[i] = 0;
        }
        else {
          vengefulness[i] = 1;
        }
      }
    }

    for (int i = 0; i < this.metaVengefulness.length; i++) {
      if (Uniform.staticNextDoubleFromTo(0, 1) <= rateMutation) { //Peligro de error de punto flotante
        if (metaVengefulness[i] == 1) {
          metaVengefulness[i] = 0;
        }
        else {
          metaVengefulness[i] = 1;
        }
      }
    }

    boldnessDecimal = setBitValuesInDecimalValues(boldness);
    vengefulnessDecimal = setBitValuesInDecimalValues(vengefulness);
    metaVengefulnessDecimal = setBitValuesInDecimalValues(metaVengefulness);

  }

  /**
   * Defection or not decision
   */
  public void defection() {

    this.variable_S = Uniform.staticNextDoubleFromTo(0,
        Math.pow(2, this.arrayDimension) - 1);

    if (this.variable_S < this.boldnessDecimal) {
      this.defection = true;
    }
    else {
      this.defection = false;
    }
  }

  /**
   *
   * @param vector int[]
   * @param dimension int
   */
  public void setRandomValuesInABitVector(int[] vector, int dimension) {

    for (int i = 0; i < dimension; i++) {
      vector[i] = Uniform.staticNextIntFromTo(0, 1);

    }
  }

  /**
   *
   * @param value double
   * @param vector int[]
   * @param dimension int
   */
  public void setFixedValuesInABitVector(double value, int[] vector,
                                         int dimension) {

    int k = (int) value;

    for (int i = dimension - 1; i >= 0; i--) {
      if (k >= (int) Math.pow(2, i)) {
        if (k == (int) Math.pow(2, i)) {
          vector[i] = 1;
          k = 0;
        }
        else {
          vector[i] = 1;
          k = k - (int) Math.pow(2, i);
        }
      }
      else {
        vector[i] = 0;
      }
    }
  }

  /**
   *
   * @return boolean
   */
  public boolean getDefection() {
    return this.defection;
  }

  /**
   *
   * @return double
   */
  public double getVariable_S() {
    return this.variable_S;
  }

  /**
   *
   * @return int
   */
  public int getVengefulnessDecimal() {
    return this.vengefulnessDecimal;
  }

  /**
   *
   * @return int
   */
  public int getBoldnessDecimal() {
    return this.boldnessDecimal;
  }

  /**
   *
   * @return int
   */
  public int getMetaVengefulnessDecimal() {
    return this.metaVengefulnessDecimal;
  }

  /**
   *
   * @param vector int[]
   * @return int
   */
  public int setBitValuesInDecimalValues(int[] vector) {

    int value = 0;
    for (int i = 0; i < vector.length; i++) {
      value = (int) (vector[i] * Math.pow(2, i) + value);
    }

    return value;
  }

}
